@@ -189,6 +189,7 @@ handle_start_element (GMarkupParseContext *context,
 	}	
 	parser->output->connected = FALSE;
 	parser->output->on = FALSE;
+	parser->output->primary = FALSE;
     }
     else if (strcmp (name, "configuration") == 0)
     {
@@ -344,6 +345,13 @@ handle_text (GMarkupParseContext *context,
 	    parser->output->rotation |= GNOME_RR_REFLECT_Y;
 	}
     }
+    else if (stack_is (parser, "primary", "output", "configuration", TOPLEVEL_ELEMENT, NULL))
+    {
+	if (strcmp (text, "yes") == 0)
+	{
+	    parser->output->primary = TRUE;
+	}
+    }
     else
     {
 	/* Ignore other properties so we can expand the format in the future */
@@ -555,6 +563,8 @@ gnome_rr_config_new_current (GnomeRRScreen *screen)
 		output->pref_height = 768;
 	    }
 	}
+
+        output->primary = gnome_rr_output_get_is_primary (rr_output);
  
 	g_ptr_array_add (a, output);
     }
@@ -985,6 +995,8 @@ emit_configuration (GnomeRRConfig *config,
 		string, "          <reflect_x>%s</reflect_x>\n", get_reflect_x (output->rotation));
 	    g_string_append_printf (
 		string, "          <reflect_y>%s</reflect_y>\n", get_reflect_y (output->rotation));
+            g_string_append_printf (
+                string, "          <primary>%s</primary>\n", yes_no (output->primary));
 	}
 	
 	g_string_append_printf (string, "      </output>\n");
@@ -998,6 +1010,7 @@ gnome_rr_config_sanitize (GnomeRRConfig *config)
 {
     int i;
     int x_offset, y_offset;
+    gboolean found;
 
     /* Offset everything by the top/left-most coordinate to
      * make sure the configuration starts at (0, 0)
@@ -1024,6 +1037,23 @@ gnome_rr_config_sanitize (GnomeRRConfig *config)
 	    output->y -= y_offset;
 	}
     }
+
+    /* Only one primary, please */
+    found = FALSE;
+    for (i = 0; config->outputs[i]; ++i)
+    {
+        if (config->outputs[i]->primary)
+        {
+            if (found)
+            {
+                config->outputs[i]->primary = FALSE;
+            }
+            else
+            {
+                found = TRUE;
+            }
+        }
+    }
 }
 
 
@@ -1151,6 +1181,7 @@ gnome_rr_config_new_stored (GnomeRRScreen *screen, GError **error)
     return config;
 }
 
+#ifndef GNOME_DISABLE_DEPRECATED_SOURCE
 gboolean
 gnome_rr_config_apply (GnomeRRConfig *config,
 		       GnomeRRScreen *screen,
@@ -1158,6 +1189,7 @@ gnome_rr_config_apply (GnomeRRConfig *config,
 {
     return gnome_rr_config_apply_with_time (config, screen, GDK_CURRENT_TIME, error);
 }
+#endif
 
 gboolean
 gnome_rr_config_apply_with_time (GnomeRRConfig *config,
@@ -1179,7 +1211,7 @@ gnome_rr_config_apply_with_time (GnomeRRConfig *config,
     {
 	if (crtc_assignment_apply (assignment, timestamp, error))
 	    result = TRUE;
-	    
+
 	crtc_assignment_free (assignment);
 
 	gdk_flush ();
@@ -1188,6 +1220,7 @@ gnome_rr_config_apply_with_time (GnomeRRConfig *config,
     return result;
 }
 
+#ifndef GNOME_DISABLE_DEPRECATED_SOURCE
 /**
  * gnome_rr_config_apply_stored:
  * @screen: A #GnomeRRScreen
@@ -1212,7 +1245,9 @@ gnome_rr_config_apply_stored (GnomeRRScreen *screen, GError **error)
 
     return result;
 }
+#endif
 
+#ifndef GNOME_DISABLE_DEPRECATED_SOURCE
 /* gnome_rr_config_apply_from_filename:
  * @screen: A #GnomeRRScreen
  * @filename: Path of the file to look in for stored RANDR configurations.
@@ -1251,6 +1286,7 @@ gnome_rr_config_apply_from_filename (GnomeRRScreen *screen, const char *filename
 {
     return gnome_rr_config_apply_from_filename_with_time (screen, filename, GDK_CURRENT_TIME, error);
 }
+#endif
 
 /* gnome_rr_config_apply_from_filename_with_time:
  * @screen: A #GnomeRRScreen
@@ -1340,6 +1376,7 @@ struct CrtcAssignment
 {
     GnomeRRScreen *screen;
     GHashTable *info;
+    GnomeRROutput *primary;
 };
 
 static gboolean
@@ -1366,6 +1403,7 @@ crtc_assignment_assign (CrtcAssignment   *assign,
 			int               x,
 			int               y,
 			GnomeRRRotation   rotation,
+                        gboolean          primary,
 			GnomeRROutput    *output)
 {
     CrtcInfo *info = g_hash_table_lookup (assign->info, crtc);
@@ -1387,6 +1425,11 @@ crtc_assignment_assign (CrtcAssignment   *assign,
 	{
 	    g_ptr_array_add (info->outputs, output);
 
+            if (primary && !assign->primary)
+            {
+                assign->primary = output;
+            }
+
 	    return TRUE;
 	}
     }
@@ -1404,6 +1447,11 @@ crtc_assignment_assign (CrtcAssignment   *assign,
 	
 	g_hash_table_insert (assign->info, crtc, info);
 	    
+        if (primary && !assign->primary)
+        {
+            assign->primary = output;
+        }
+
 	return TRUE;
     }
     
@@ -1421,6 +1469,11 @@ crtc_assignment_unassign (CrtcAssignment *assign,
     {
 	g_ptr_array_remove (info->outputs, output);
 
+        if (assign->primary == output)
+        {
+            assign->primary = NULL;
+        }
+
 	if (info->outputs->len == 0)
 	    g_hash_table_remove (assign->info, crtc);
     }
@@ -1541,6 +1594,7 @@ real_assign_crtcs (GnomeRRScreen *screen,
 			    assignment, crtc, modes[j],
 			    output->x, output->y,
 			    output->rotation,
+                            output->primary,
 			    gnome_rr_output))
 		    {
 			if (real_assign_crtcs (screen, outputs + 1, assignment))
@@ -1612,14 +1666,18 @@ crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs, GError **
     {
 	int width, height;
 	int min_width, max_width, min_height, max_height;
+	int required_pixels, min_pixels, max_pixels;
 
 	get_required_virtual_size (assignment, &width, &height);
 
 	gnome_rr_screen_get_ranges (
 	    screen, &min_width, &max_width, &min_height, &max_height);
-    
-	if (width < min_width || width > max_width ||
-	    height < min_height || height > max_height)
+
+	required_pixels = width * height;
+	min_pixels = min_width * min_height;
+	max_pixels = max_width * max_height;
+
+	if (required_pixels < min_pixels || required_pixels > max_pixels)
 	{
 	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
 			 /* Translators: the "requested", "minimum", and
@@ -1679,7 +1737,7 @@ crtc_assignment_apply (CrtcAssignment *assign, guint32 timestamp, GError **error
      */
 
     gdk_x11_display_grab (gdk_screen_get_display (assign->screen->gdk_screen));
-    
+
     /* Turn off all crtcs that are currently displaying outside the new screen,
      * or are not used in the new setup
      */
@@ -1696,7 +1754,7 @@ crtc_assignment_apply (CrtcAssignment *assign, guint32 timestamp, GError **error
 
 	    w = gnome_rr_mode_get_width (mode);
 	    h = gnome_rr_mode_get_height (mode);
-	    
+
 	    if (crtc_is_rotated (crtc))
 	    {
 		int tmp = h;
@@ -1740,6 +1798,8 @@ crtc_assignment_apply (CrtcAssignment *assign, guint32 timestamp, GError **error
 	success = !state.has_error;
     }
 
+    gnome_rr_screen_set_primary_output (assign->screen, assign->primary);
+
     gdk_x11_display_ungrab (gdk_screen_get_display (assign->screen->gdk_screen));
 
     return success;
